"""
Class composition of all the other classes, to access the
wrappers with syntactic sugar.
"""
from typing import Optional, TypeVar

from dfi.connect import Connect
from dfi.services.datasets import Datasets
from dfi.services.delete import Delete
from dfi.services.get import Get
from dfi.services.info import Info
from dfi.services.ingest import Ingest
from dfi.services.users import Users

TClient = TypeVar("TClient", bound="Client")


class Client:
    """
    Client class gathering all the classes and method in a single one. Facade of dfi.Connect.

    See documentation of dfi.Connect for its input values.

    :example:
    ```python
    from dfi import Client

    dfi = Client(token, url)
    ```
    """

    def __init__(
        self,
        api_token: str,
        base_url: Optional[str] = None,
        query_timeout: Optional[int] = 60,
        progress_bar: Optional[bool] = False,
    ) -> None:
        self.conn = Connect(
            api_token=api_token,
            base_url=base_url,
            query_timeout=query_timeout,
            progress_bar=progress_bar,
        )
        self.datasets = Datasets(self.conn)
        self.delete = Delete(self.conn)
        self.get = Get(self.conn)
        self.info = Info(self.conn)
        self.ingest = Ingest(self.conn)
        self.users = Users(self.conn)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""



import json
from datetime import datetime
from enum import Enum
from typing import Optional

from dfi import models, validate


class FilterOperator(str, Enum):
    lt = "lt"
    gt = "gt"
    gte = "gte"
    lte = "lte"
    eq = "eq"
    neq = "neq"
    between = "between"
    outside = "outside"


FILTERS = "filters"
TIME = "time"
MIN_TIME = "minTime"
MAX_TIME = "maxTime"
RETURN_MODEL = "return"
COORDINATES = "coordinates"
FILTER_FIELDS = "fields"
DATASET_ID = "datasetId"
GEO = "geo"


class QueryDocument(dict):

    """A high level interface for declaring a DFI API Query"""

    def __init__(self) -> None:
        self._internal_document = {FILTERS: {}, RETURN_MODEL: "records"}

    def set_dataset_id(self, id: str) -> None:
        self._internal_document[DATASET_ID] = id

    def get_dataset_id(self) -> Optional[str]:
        return self._internal_document.get(DATASET_ID)

    def set_time_range(self, time_interval: models.TimeInterval):
        validate.time_interval(time_interv=time_interval)
        if not self._internal_document.get(TIME):
            self._internal_document[FILTERS][TIME] = {}
        self._internal_document[FILTERS][TIME][MIN_TIME] = self._format_timestamp(time_interval[0])
        self._internal_document[FILTERS][TIME][MAX_TIME] = self._format_timestamp(time_interval[1])

    def get_min_time(self) -> Optional[datetime]:
        return self._internal_document[FILTERS].get(TIME).get(MIN_TIME)

    def get_max_time(self) -> Optional[datetime]:
        return self._internal_document[FILTERS].get(TIME).get(MAX_TIME)

    @staticmethod
    def _format_timestamp(dt: datetime) -> str:
        """
        Uses formatting in unpack_and_convert_time_interval under services.get
        """
        return dt.strftime("%Y-%m-%dT%H:%M:%S.%fZ")

    def set_polygon(self, vertices: models.Polygon) -> None:
        if isinstance(vertices[0], float):
            # The polygon is a bounding box
            raise ValueError("Bounding boxes are not able to be coerced to polygons")

        validate.polygon(poly=vertices)
        self._internal_document[FILTERS][GEO] = {"type": "Polygon", COORDINATES: vertices}

    def get_polygon(self) -> Optional[models.Polygon]:
        return self._internal_document[FILTERS].get(GEO).get(COORDINATES)

    def set_id_filter(self, ids: list) -> None:
        validate.entities(input_entities=ids)
        self._internal_document[FILTERS]["id"] = ids

    def add_filter_field(self, field_name: str, operation: FilterOperator, value) -> None:
        if not self._internal_document[FILTERS].get(FILTER_FIELDS):
            self._internal_document[FILTERS][FILTER_FIELDS] = dict()
        if list(FilterOperator).__contains__(operation):
            self._internal_document[FILTERS][FILTER_FIELDS][field_name] = {operation: value}
        else:
            raise validate.DFIInputValueError(f"{operation} not a valid filter operation")

    def add_filter_fields_from_dict(self, fields_dict: models.FilterFields) -> None:
        """
        Set multiple filter fields by passing in a dictionary with format
        {FIELD_NAME: {FILTER_OPERATION: VALUE}}
        """
        if not self._internal_document[FILTERS].get(FILTER_FIELDS):
            self._internal_document[FILTERS][FILTER_FIELDS] = dict()

        for k, v in fields_dict.items():
            if isinstance(v, dict):
                for k_2, v_2 in v.items():
                    self.add_filter_field(k, k_2, v_2)
            else:
                raise validate.DFIInputValueError()

    def get_document(self) -> dict:
        """
        Returns a query document that can be used to submit a query
        """
        return self._internal_document

    def filter_fields_url_encoding(self) -> dict:
        """
        While we are in a transition period we still need to encode filter field parameters as a URL parameter

        These can be tacked onto other stuff going into query params.
        """

        fields: dict = self._internal_document[FILTERS].get(FILTER_FIELDS, dict())
        query_fields = dict()

        for k, v in fields.items():
            for k_2, v_2 in v.items():
                query_fields[f"{k}[{k_2}]"] = json.dumps(v_2)

        return query_fields



"""
Manages user authentication via the Users API service.
"""
from typing import List

from dfi.connect import Connect


class Users:
    """
    Class responsible handling users, identities, and tokens.

    It can be accessed via the a dfi.Client class instance or it must be instantiated
    with a dfi.Connect instance as argument.
    """

    def __init__(self, conn: Connect) -> None:
        self.conn = conn

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""

    def get_tokens(self, include_expired: bool = False) -> List[dict]:
        """
        Retrieve details about API tokens

        :param include_expired: set to `True` to include expired tokens in the list.  Defaults to False.
        :return: a list of API tokens.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)
        tokens = dfi.users.get_tokens(include_expired=False)
        ```
        """
        params = {"includeExpired": str(include_expired).lower()}
        with self.conn.api_get("v1/tokens", params=params, stream=False) as response:
            response.raise_for_status()
            return response.json()

    def create_token(self, name: str, validity: str) -> dict:
        """
        Generate a new API token for the current user.

        :param name: name for token.
        :param validity: how long the token is valid for.  Period of token validity in ISO8601 format. Default is 1 year.

            pattern: `^P(?:(\d+)Y)?(?:(\d+)M)?(?:(\d+)D)?(?:T(?:(\d+)H)?(?:(\d+)M)?(?:(\d+(?:\.\d{1,3})?)S)?)?$`
            example: `"P1Y"`


        :return: information about the new token.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        name = "token_1"
        validity = "P1Y"
        token_info = dfi.users.create_token(name, validity)
        token = token_info["token]
        ```
        """
        payload = {"name": name, "validity": validity}
        with self.conn.api_post("v1/tokens", payload=payload, stream=False) as response:
            response.raise_for_status()
            return response.json()

    def expire_token(self, token_id: str) -> None:
        """
        Expires the given api token.

        :::{note}
        It's previous existence can still be found with `dfi.get_tokens(include_expired=True)`
        :::

        :param token_id: id of token to expire.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        dfi.users.expire_token()
        ```
        """
        with self.conn.api_delete(f"v1/tokens/{token_id}", stream=False) as response:
            response.raise_for_status()
            return None

    def get_identities(self) -> List[dict]:
        """
        Retrieve a list of identities.

        :::{hint}
        You need to be an admin for this request.
        :::

        :return: None
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        identities = dfi.users.get_identities()
        ```
        """
        with self.conn.api_get("v1/identities", stream=False) as response:
            response.raise_for_status()
            return response.json()

    def get_my_identity(self) -> dict:
        """
        Retrieve data about the identity who made this request

        :return: None
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        identity = dfi.users.get_my_identity()
        ```
        """
        with self.conn.api_get("v1/identities/me", stream=False) as response:
            response.raise_for_status()
            return response.json()

    def get_identity(self, identity_id: str) -> dict:
        """
        Retrieve data about a specific identity.

        :::{hint}
        You need to be an admin for this request.
        :::

        :param identity_id: an id of an identity.
        :return: the identity of the given id.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        identity = dfi.users.get_identity("test_id")
        ```
        """
        with self.conn.api_get(f"v1/identities/{identity_id}", stream=False) as response:
            response.raise_for_status()
            return response.json()

    def _delete_identity(self, identity_id: str) -> None:
        """
        Remove an identity and all opaque tokens.

        :::{attention}
        This should be done in conjunction with removing the user.
        :::

        :::{hint}
        You need to be an admin for this request.
        :::

        :::{warning}
        Deleting the sole identity for a user makes the user unsearchable,
        which makes finding the id to then delete the user impossible.
        :::

        :param identity_id: an id of an identity.
        :return: None
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        identity = dfi.users.delete_identity("user|xxxyyyzzz)
        ```
        """
        with self.conn.api_delete(f"v1/identities/{identity_id}", stream=False) as response:
            response.raise_for_status()
            return None

    def create_user(self, user: dict) -> dict:
        """
        Create a new user.

        :::{hint}
        You need to be an admin for this request.
        :::

        :param user: a user specification to create.
        :return: the created user.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        user = {
            "userName": "iain.banks",
            "name": {
                "givenName": "Iain",
                "familyName": "Banks",
            },
            "emails": [
                {
                    "value": "iain.banks@organization.com",
                    "primary": True
                }
            ],
            "active": True,
        }
        created_user = dfi.users.create_user(user)
        ```
        """
        with self.conn.api_post("v1/users", payload=user, stream=False) as response:
            response.raise_for_status()
            return response.json()

    def get_user(self, user_id: str) -> None:
        """
        Retrieve details about a user by id.

        :::{hint}
        You need to be an admin for this request.
        :::

        :param user_id: an id of a user.
        :return: the identity of the given id.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        identity = dfi.users.get_user("user_123")
        ```
        """
        with self.conn.api_get(f"v1/users/{user_id}", stream=False) as response:
            response.raise_for_status()
            return response.json()

    def delete_user(self, user_id: str) -> None:
        """
        Removes a user from the system. This removes all identities and access tokens that exist for the user as well.

        :::{hint}
        You need to be an admin for this request.
        :::

        :param user_id: an id of a user.
        :return: the identity of the given id.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        identity = dfi.users.delete_user("user_123")
        ```
        """
        with self.conn.api_delete(f"v1/users/{user_id}", stream=False) as response:
            response.raise_for_status()
            return None



"""
Class with DFI getters, wrappers of the DFI python API for ingest methods.
Composition of the class Connection.
"""
import logging
from typing import List, Optional, TypeVar, Union

from dfi import validate
from dfi.connect import Connect

TIngest = TypeVar("TIngest", bound="Ingest")

_logger = logging.getLogger(__name__)


class Ingest:
    """
    Class responsible to call the HTTP API and ingest data into the DFI.

    It can be accessed via the a dfi.Client class instance or it must be instantiated
    with a dfi.Connect instance as argument.

    :param conn: Instance of a Connect.
    :example:
    ```python
    from dfi import Client

    dfi = Client(token, dataset_id, url)
    dfi.ingest.from_file(file_to_spatiotemporal_data)
    ```
    """

    def __init__(self, conn: Connect) -> None:
        self.conn = conn
        self.dataset_id: Optional[str]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""

    def from_data(self, dataset_id: str, data: List[dict]) -> dict:
        """
        Insert data in an instance of a Data Flow Index, passed as a list of dicts, with coordinates.

        :param dataset_id: the dataset to ingest data into.
        :param data: List of dict as in the example below.
        :raises `DFIInputDataError`: if the input dictionaries do not have the correct keys.
        :raises `DFIInputValueOutOfBoundError`: if the input coordinates are out of bound.
        :return: dict with a status message.
        :example:
        ```python
        from dfi import Client

        dataset_id = "1234"
        data = [
            {
                "coordinate": [-73.985664,40.748441],
                "time": "2022-09-01T17:32:28.250Z",
                "id": 0,
                "payload": "Application specific data"
            },
            {
                "coordinate": [-73.985645,40.748423],
                "time": "2022-09-01T17:32:40.310Z",
                "id": 1,
                "payload": "Application specific data"
            }
        ]

        dfi = Client(token, url)

        dfi.insert.from_data(dataset_id, data)
        ```
        """
        validate.data(data)

        params = {"instance": dataset_id}

        with self.conn.api_post("insert", params=params, payload=data) as response:
            msg = f"Ingesting data. Response from DFI {response.text}"
            _logger.info(msg)
            result = response.json()
            result.update({"msg": msg})
            return result

    def from_s3(self, dataset_id: str, url_s3: Union[str, List[str]], data_format: Optional[dict] = None) -> dict:
        """
        Insert data in an instance of a Data Flow Index from a pre-signed URL, or a list of pre-signed URLs
        to an s3 bucket the user have access to, and can create pre-signed URLs.

        A pre-signed URL can be created from the AWS UI selecting a file from a bucket, then selecting "share with a presigned URL"
        under the "Action" drop-bar.

        :param dataset_id: the dataset to ingest data into.
        :param url_s3: pre-signed S3 URL pointing to a file, or a list of pre-signed S3 URL pointing to files.
        :param data_format: Optional dict where the columns are specified. If None it will use the default:
            ```python
            data_format = {
                "columns":
                    {
                        "entityId": 0,
                        "timestamp": 1,
                        "longitude": 2,
                        "latitude": 3
                    }
                }
            ```
        :return: dict with a status message.
        :raises `DFIInputValueError`: if the input is not a string or a list of strings.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)
        data_format = {
            "columns":
                {
                    "entityId": 0,
                    "timestamp": 1,
                    "longitude": 2,
                    "latitude": 3
                }
            }
        dfi.insert.from_s3(dataset_id, <path to s3 bucket>, data_format=data_format)
        ```
        """
        validate.url_s3(url_s3)

        if data_format is None:
            data_format = {"columns": {"entityId": 0, "timestamp": 1, "longitude": 2, "latitude": 3}}

        urls = url_s3
        if isinstance(url_s3, str):
            urls = [urls]

        payload = {
            "instance": dataset_id,
            "source": {"urls": urls},
            "format": data_format,
        }
        with self.conn.api_put("import/batch", params=payload) as response:
            msg = f"Ingesting data from s3 URL {url_s3}. Response from DFI {response.text}"
            _logger.info(msg)
            result = response.json()
            result.update({"msg": msg})
            return result

    def check_s3_import_status(self, ingestion_id: str) -> dict:
        """
        Check the status of an ingestion via s3.

        :param ingestion_id: str with the ingestion id obtained as the output of self.from_s3.
        :return: dict with a status message and the count of entities inserted.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, dataset_id, url)
        dfi.insert.check_s3_import_status(<ingestion id>)
        ```
        """
        with self.conn.api_get(f"import/batch/{ingestion_id}/status") as response:
            msg = f"Status for ingestion {ingestion_id} request submitted. Response from DFI {response.text}"
            _logger.info(msg)
            result = response.json()
            result.update({"msg": msg})
            return result



"""
Manages queries to the Datasets API.  Allows users to find, create, update, and delete
datasets, dataset schemas, and dataset permissions.
"""
import logging
from datetime import datetime
from typing import List, Optional

import requests

from dfi.connect import Connect

_logger = logging.getLogger(__name__)


class Datasets:
    """
    Class responsible handling datasets and dataset schemas.

    It can be accessed via the a dfi.Client class instance or it must be instantiated
    with a dfi.Connect instance as argument.
    ```
    """

    def __init__(self, conn: Connect) -> None:
        self.conn = conn

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""

    def create(self, dataset: dict) -> dict:
        """
        Create an empty dataset.

        :param dataset: the dataset definition to be created.
        :return: the added dataset definition.  The `dataset_id` is returned under the `id` key.
        :example:
        ```python
        from dfi import Client

        dataset = {
          "name": "test-0",
          "description": None,
          "active": True,
          "tags": {},
          "type": "managed",
          "model": "point",
          "dataDescription": {
            "metadataSchema": {
              "plantHeight": {
                "type": "number",
                "nullable": False,
                "signed": False
              },
              "plantIPv4": {
                "type": "ip",
                "nullable": True
              },
              "plantCultivar": {
                "type": "enum",
                "nullable": True,
                "values": [
                  "broccoli",
                  "brocollini",
                  "brussel sprouts",
                  "cabbage",
                  "cauliflower",
                  "collards"
                ]
              }
            },
            "boundingBox": [-180.0, -90.0, 180.0, 90.0],
            "minDatetime": "2021-01-01T00:00:00.000Z",
            "maxDatetime": "2022-01-01T00:00:00.000Z"
          },
          "source": {
            "s3SourceUrl": "s3://test-bucket/dataset-0"
          },
          "pipeline": {
            "curationConfiguration": {}
          },
          "pii": {
            "keepPii": False,
            "piiFields": []
          },
          "storage": {
            "dataStoreType": "dfi",
            "dataStoreConnectionDetails": {
              "host": "0.0.0.0",
              "port": "1234",
              "queryTimeout": "3600000"
            }
          },
          "destination": {
            "dataStoreRetentionLength": 0
          },
          "permissions": [
            {
              "type": "reader",
              "scope": "all"
            }
          ]
        }

        dfi = Client(token, url)
        dfi.datasets.create(dataset)
        ```
        """
        with self.conn.api_post("v1/datasets", stream=False, payload=dataset) as response:
            response.raise_for_status()
            return response.json()

    def find(
        self, name: Optional[str] = None, before: Optional[datetime] = None, limit: Optional[int] = None
    ) -> List[dict]:
        """
        Find datasets.

        :param name: retrieves all datasets with `name`.
        :param before: retrieves all datasets created before `before`.
        :param limit: only return `limit` number of datasets.
        :return: a list of dataset definitions.
        :example:
        ```python
        from datetime import datetime
        from dfi import Client

        dfi = Client(token, url)

        time = datetime.fromisoformat("2023-03-31T12:50:00Z")
        dfi.datasets.find(name="London", before=time, limit=10)
        ```
        """
        if before:
            before = before.isoformat()

        params = {"name": name, "before": before, "limit": limit}

        with self.conn.api_get("v1/datasets", stream=False, params=params) as response:
            response.raise_for_status()
            return response.json()

    def find_by_id(self, dataset_id: str) -> dict:
        """
        Finds a dataset by id.

        :param dataset_id: id of the dataset to retrieve.
        :return: dataset definition.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)
        dfi.datasets.find_by_id(dataset_id="1234")
        ```
        """
        with self.conn.api_get(f"v1/datasets/{dataset_id}", stream=False) as response:
            response.raise_for_status()
            return response.json()

    def update(self, dataset_id: str, dataset: dict) -> dict:
        """
        Update a dataset by id.

        :param dataset_id: id of the dataset to update.
        :param dataset: new dataset definition.
        :return: updated dataset definition.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        dataset_id = "1234"

        dataset_update = {
            "description": "a test dataset",
            "dataDescription": {
                "minDatetime": "2020-01-01T00:00:00.000Z",
                "maxDatetime": "2021-01-01T00:00:00.000Z",
            },
        }

        dfi.datasets.update(dataset_id, dataset_update)
        ```
        """
        with self.conn.api_patch(f"v1/datasets/{dataset_id}", stream=False, payload=dataset) as response:
            response.raise_for_status()
            return response.json()

    def delete(self, dataset_id: str) -> None:
        """
        Delete a dataset by id.

        :param dataset_id: id of the dataset to delete.
        :return: None
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        dataset_id = "1234"
        dfi.datasets.delete(dataset_id)
        ```
        """
        with self.conn.api_delete(f"v1/datasets/{dataset_id}", stream=False) as response:
            response.raise_for_status()
            return None

    def get_permissions(self, dataset_id: str) -> dict:
        """
        Get list of permissions for this dataset.

        :param dataset_id: id of the dataset to retrieve permissions from.
        :return: list of permissions for dataset.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)
        dfi.datasets.get_permissions(dataset_id="1234")
        ```
        """
        with self.conn.api_get(f"v1/datasets/{dataset_id}/permissions", stream=False) as response:
            response.raise_for_status()
            return response.json()

    def add_permissions(self, dataset_id: str, permissions: List[dict]) -> dict:
        """
        Add new permissions to a dataset.

        :param dataset_id: id of the dataset to add permissions to.
        :param permissions: list of permissions to be added.
        :return: list of added permissions for the dataset.
        :example:

        ## I. Update permissions so anyone can read the dataset
        ```python
        from dfi import Client

        dataset_id = "1234"
        permissions = [
            {
                "type": "reader",
                "scope": "all"
            }
        ]
        dfi = Client(token, url)
        updated_permissions = dfi.datasets.add_permissions(dataset_id=dataset_id, permissions=permissions)
        ```

        ## II. Update permissions specific user has write permissions
        ```python
        from dfi import Client

        dataset_id = "1234"
        permissions = [
            {
                "type": "writer",
                "scope": "identity",
                "identityId": "user-123"
            }
        ]
        dfi = Client(token, url)
        updated_permissions = dfi.datasets.add_permissions(dataset_id=dataset_id, permissions=permissions)
        ```
        """
        with self.conn.api_post(f"v1/datasets/{dataset_id}/permissions", stream=False, payload=permissions) as response:
            response.raise_for_status()
            return response.json()

    def delete_permissions(self, dataset_id: str, permissions: List[dict]) -> dict:
        """
        Remove permissions from a dataset.  Given permission must match exactly to be removed.

        :param dataset_id: id of the dataset to remove permissions from.
        :param permissions: permissions to be deleted from the dataset.
        :return: list of deleted permissions from the dataset.
        :example:
        ```python
        from dfi import Client

        dataset_id = "1234"
        permissions = [{"type": "reader", "scope": "identity", "identityId": "123"}, {"type": "reader", "scope": "all"}]
        dfi = Client(token, url)
        dfi.datasets.delete_permissions(dataset_id=dataset_id, permissions=permissions)
        ```
        """
        with self.conn.api_delete(
            f"v1/datasets/{dataset_id}/permissions", stream=False, payload=permissions
        ) as response:
            response.raise_for_status()
            return response.json()

    def get_user_permissions(self, dataset_id: str) -> dict:
        """
        Get list of current user permissions on a dataset.

        :param dataset_id: id of the dataset to retrieve user permissions from.
        :return: a list of user permissions for the dataset.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)
        dfi.datasets.get_user_permissions(dataset_id="1234")
        ```
        """
        with self.conn.api_get(f"v1/datasets/{dataset_id}/permissions/me", stream=False) as response:
            response.raise_for_status()
            return response.json()

    def get_schema_by_id(self, dataset_id: str, schema_type: str = "full", media_type: str = "json") -> dict:
        """
        Retrieve a copy of the schema for this dataset.

        :param dataset_id: id of the dataset to retrieve schema from.
        :param schema_type: set which type of Feather schema to return.  Defaults to "full".
            - `full` - Return a full Feather schema with all fields
            - `core` - Return a Feather schema with only the core fields
            - `withMetadataId` - return a Feather schema with the core fields + metadataId field.
            - `withFilterFields` - return a Feather schema with the core fields + filter fields.
        :param media_type: (`json` or `feather`) defines the file format for the returned schema, either `"json"`
            or `"feather"`.  Defaults to `"json"`.
        :return: the dataset's schema.
        :example:

        # Retrieve dataset schema as JSON
        ```python
        from dfi import Client

        dfi = Client(token, url)
        schema_as_json = dfi.datasets.get_schema_by_id(dataset_id="1234", schema_type="full", media_type="json")
        ```

        # Retrieve dataset schema as Feather bytes
        ```python
        from dfi import Client

        dfi = Client(token, url)
        schema_as_json = dfi.datasets.get_schema_by_id(dataset_id="1234", schema_type="full", media_type="feather")
        """
        schema_types = {
            "full": "full",
            "core": "core",
            "with_metadata_id": "withMetadataId",
            "with_filter_fields": "withFilterFields",
        }
        params = {"type": schema_types[schema_type]}

        headers = {
            "Authorization": f"Bearer {self.conn.api_token}",
            "accept": "application/json",
        }
        headers["content-type"] = "application/feather" if media_type == "feather" else "application/json"
        url = f"{self.conn.base_url}/v1/datasets/{dataset_id}/schema"
        _logger.debug(
            {"url": url, "headers": headers, "stream": False, "params": params, "timeout": self.conn.query_timeout}
        )

        with requests.get(
            url,
            headers=headers,
            stream=False,
            params=params,
            timeout=self.conn.query_timeout,
        ) as response:
            response.raise_for_status()

            return response.content if media_type == "feather" else response.json()

    def add_enums(self, dataset_id: str, metadata_enums: dict) -> dict:
        """
        Add new values to an enum field. Any new values added to fields here are merged into the existing values.

        :param dataset_id: id of the dataset to add enum values to.
        :param metadata_enums: a dictionary of metadata fields with enums values to be added to the dataset schema.
        :return: the newly updated metadata schema
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        metadata_enums = {
            "plantCultivar": {"type": "enum", "values": ["kale", "kohlrabi", "mustard"], "nullable": True},
        }
        dfi.datasets.add_enums(dataset_id="1234", metadata_enums=metadata_enums)
        ```
        """
        with self.conn.api_post(
            f"v1/datasets/{dataset_id}/schema/values", stream=False, payload=metadata_enums
        ) as response:
            response.raise_for_status()
            return response.json()



"""
Class with DFI getters, wrappers of the DFI python API for info methods.
Composition of the class Connection.
"""
import logging
from typing import List, Union

import requests

from dfi import __version__ as version
from dfi import validate
from dfi.connect import Connect

_logger = logging.getLogger(__name__)


class Info:
    """
    Class responsible to call the HTTP API and submit queries for information methods.
    ```
    """

    def __init__(self, conn: Connect) -> None:
        self.conn = conn

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""

    def api_version(self) -> str:
        """
        Returns the version of the Data Flow Index (DFI) API.
        """
        with self.conn.api_get("version", stream=False, params=None) as response:
            response.raise_for_status()
            return response.text

    def dfipy_version(self) -> str:
        """
        Returns the version of the dfipy python library.
        """
        return version

    def product_version(self) -> str:
        """
        Returns the Data Flow Index (DFI) product version.
        """
        with self.conn.api_get("product/version", stream=False, params=None) as response:
            try:
                return response.json()["version"]
            except KeyError as err:
                raise validate.DFIResponseError(
                    "Product version endpoint should return a dictionary with 'version' as key. "
                    f"It returned instead {response.json()}"
                ) from err

    def instances(self) -> List[dict]:
        """
        Returns the list of available instances serviced by the Data Flow Index.

        :returns: List of dictionaries containing the names and descriptions of rhe DFI instances available
            for the given url.
        :raises `DFIResponseError`: if the queried json can not be parsed correctly, or if the resource is not available.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)
        dfi.get.instances()
        ```
        """
        with self.conn.api_get("instances") as response:
            return self._receive_json(response)

    def profile(self) -> dict:
        """
        Returns the details of your profile.

        :returns: a dictionary with the details referring to your user's profile. These are base on the API token you have selected.
        :raises `DFIResponseError`: if the queried json can not be parsed correctly, or if the resource is not available.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)
        dfi.get.profile()
        ```
        """
        with self.conn.api_get("profile") as response:
            return self._receive_json(response)

    @staticmethod
    def _receive_json(response: requests.models.Response) -> Union[List[dict], dict]:
        try:
            return response.json()
        except TypeError as type_err:
            msg = f"Parsing response to json returned: {type_err}"
        except requests.exceptions.HTTPError as http_err:
            msg = f"HTTP error when parsing requests json: {http_err}"
        except requests.exceptions.JSONDecodeError as decode_err:
            msg = f"Failed to parse response to json with error: {decode_err}"
        _logger.error(msg)
        raise validate.DFIResponseError(msg)



"""
Class with DFI getters, wrappers of the DFI python API for delete methods.
Composition of the class Connection.
"""
import json
import logging

from dfi.connect import Connect

_logger = logging.getLogger(__name__)


class Delete:
    """
    Class responsible to call the HTTP API and delete data from DFI.

    It can be accessed via the a dfi.Client class instance or it must be instantiated
    with a dfi.Connect instance as argument.

    :param conn: Instance of a Connect.
    :example:
    ```python
    from dfi import Client

    dfi = Client(token, url)
    dfi.delete.truncate(dataset_id)
    ```
    """

    def __init__(self, conn: Connect) -> None:
        self.conn = conn

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""

    def truncate(self, dataset_id: str) -> dict:
        """
        Delete all data in the target DFI instance.

        :::{hint}
        You need to be an admin for this request.
        :::

        :param dataset_id: id of the dataset.
        :return: dict with the response status message.
        :example:
        ```python
        from dfi import Client

        dfi = Client(dfi_token, dataset_id, base_url)
        dfi.delete.truncate(dataset_id)
        ```
        """
        with self.conn.api_post(f"instances/{dataset_id}/truncate") as response:
            msg = f"Truncating instance '{dataset_id}'. DFI Response: '{response.text}' "
            result = {"status code": response.status_code}
            _logger.info(msg)
            result.update({"msg": msg})
            try:
                result.update({"response.json": response.json()})
            except json.JSONDecodeError:
                pass

            return result



"""
Class with DFI getters, wrappers of the DFI python API.
Composition of the class Connection.
"""
import json
import logging
from datetime import datetime
from typing import Any, List, Optional, Tuple, TypeVar, Union

import pandas as pd
import requests
import sseclient
from tqdm import tqdm

from dfi import models, validate
from dfi.connect import Connect
from dfi.query_document import QueryDocument

TGet = TypeVar("TGet", bound="Get")

_logger = logging.getLogger(__name__)

NUM_ATTEMPTS = 3


class Get:
    """
    Class responsible to call the HTTP API and submit queries to get data from DFI.

    It can be accessed via the a dfi.Client class instance or it must be instantiated
    with a dfi.Connect instance as argument.

    :param conn: Instance of a Connect.
    :example:
    ```python
    from dfi import Client

    dfi = Client(token, url)

    dataset_id = "1234"
    dfi.get.entities(dataset_id, time_interval=[start_time, end_time])
    ```
    """

    def __init__(self, conn: Connect) -> None:
        self.conn = conn

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.conn!r})"

    def __str__(self) -> str:
        return f"""Instance of dfi.{self.__class__.__name__} composed with: {self.conn!s}"""

    def records_count(
        self,
        dataset_id: str,
        entities: Optional[List[str]] = None,
        polygon: Optional[models.Polygon] = None,
        time_interval: Optional[models.TimeInterval] = None,
        filter_fields: Optional[dict] = None,
    ) -> int:
        """
        Queries for the number of records within the bounds.

        If all the variables are None it returns the total number of records in the
        database. Start_time and end_time must be both valid datetime, with
        start_time < end_time or both None.

        :param dataset_id: the dataset to be queried.
        :param polyogn: List of vertices `[[lon1, lat1], [lon2, lat2], [lon3, lat3], [lon1, lat1]]` or a list of four
            floats representing the bounding box extremes as `[lon_min, lat_min, lon_max, lat_max]`.
            Non-valid input will raise an error.
        :param time_interval: Tuple with the lower bound and the upper bound time constraints.
        :param filter_fields: Optional[dict]: {field_name: {FilterOperation: value}}
        :returns: The number of records stored in the DFI engine.
        :raises `DFIInputValueError`: If `time_interval` or `polygon` are ill-formed.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        dataset_id = "1234"
        start_time = datetime.strptime("2022-01-01T08:00:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ")
        end_time = datetime.strptime("2022-01-01T08:30:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ")
        entities = ["299eb59a-e47e-48c0-9ad5-89a9ce1303f4"]

        dfi.get.records_count(
            dataset_id,
            polygon = None,
            time_interval = (start_time, end_time),
            entities = entities,
        )
        ```
        """
        query_doc = _init_query_document(
            dataset_id=dataset_id, polygon=polygon, time_interval=time_interval, filter_fields=filter_fields
        )

        params = {"instance": dataset_id}
        payload = {}

        if polygon is None and entities is None:
            if time_interval is not None:
                params["startTime"], params["endTime"] = unpack_and_convert_time_interval(time_interval)
            params.update(query_doc.filter_fields_url_encoding())
            with self.conn.api_get("count", params=params) as response:
                return self._receive("count", response)

        if polygon is None and entities is not None:
            if time_interval is not None:
                params["startTime"], params["endTime"] = unpack_and_convert_time_interval(time_interval)
            sum_records = 0
            params.update(query_doc.filter_fields_url_encoding())
            for uid in entities:
                with self.conn.api_get(f"entities/{uid}/count", params=params) as response:
                    sum_records += self._receive("count", response)
            return sum_records

        if isinstance(polygon[0], list) or isinstance(polygon[0], tuple):
            payload = {}
            if time_interval is not None:
                payload["startTime"], payload["endTime"] = unpack_and_convert_time_interval(time_interval)
            if entities is not None:
                payload["include"] = entities
            if filter_fields:
                payload["fields"] = query_doc.get_document()["filters"]["fields"]
            payload["vertices"] = polygon
            with self.conn.api_post("polygon/count", payload=payload, params=params) as response:
                return self._receive("count", response)

        if isinstance(polygon[0], float):
            if time_interval is not None:
                params["startTime"], params["endTime"] = unpack_and_convert_time_interval(time_interval)
            if entities is not None:
                params["include"] = entities

            min_lng, min_lat, max_lng, max_lat = polygon
            params.update(query_doc.filter_fields_url_encoding())
            with self.conn.api_get(
                f"bounding-box/{min_lng}/{min_lat}/{max_lng}/{max_lat}/count", params=params
            ) as response:
                return self._receive("count", response)

    def records(
        self,
        dataset_id: str,
        entities: Optional[List[str]] = None,
        polygon: Optional[models.Polygon] = None,
        time_interval: Optional[models.TimeInterval] = None,
        add_payload_as_json: bool = False,
        filter_fields: dict = None,
    ) -> pd.DataFrame:
        """
        Get the records of the entities appearing within the given time, space and entity ids constraints.

        List of entities and polygon can not be left both to None.

        Start time and end time passed in the time_interval must be both valid datetime,
        with start_time < end_time or both None.

        :param dataset_id: the dataset to be queried.
        :param polygon: List of vertices [[lon1, lat1], [lon2, lat2], ...] or a list of four
            floats representing the bounding box extremes as [lon_min, lat_min, lon_max, lat_max].
            Non-valid input will raise an error.
        :param time_interval: Tuple with the Lower bound and the upper bound time constraints.
        :param entities: List of entity ids. It must be passed as list also for a single element.
        :param add_payload_as_json: If True it parses the payload as a JSON string into the column payload.
        :param filter_fields: Optional[dict]: {field_name: {FilterOperation: value}}
        :returns: Dataframe with the records of the entities found in polygon, given the input constraints.
        :raises `DFIInputValueError`: If `time_interval` or `polygon` are ill-formed.
        :raises `ValueError`: If no filter bound is specified.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        dataset_id = "1234"
        start_time = datetime.strptime("2022-01-01T08:00:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ")
        end_time = datetime.strptime("2022-01-01T08:30:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ")
        entities = ["299eb59a-e47e-48c0-9ad5-89a9ce1303f4"]

        dfi.get.records(
            dataset_id,
            time_interval = (start_time, end_time),
            entities = entities,
            add_payload_as_json=True
        )

        ```
        """
        query_doc = _init_query_document(
            dataset_id=dataset_id, polygon=polygon, time_interval=time_interval, filter_fields=filter_fields
        )

        params = {"instance": dataset_id}
        payload = {}

        if polygon is None:
            if entities is None:
                raise ValueError("You have to pass a list of entity ids or a polygon, or both.")
            if time_interval is not None:
                params["startTime"], params["endTime"] = unpack_and_convert_time_interval(time_interval)

            df_entities = []
            params.update(query_doc.filter_fields_url_encoding())
            for uid in entities:
                with self.conn.api_get(f"entities/{uid}/history", params=params) as response:
                    data = self._receive("history", response)
                    _logger.debug("Uid: %s \nHistory length: %i", uid, len(data))
                    data_formatted = []

                    for item in data:
                        payload = {}
                        if add_payload_as_json:
                            _logger.warning("Payload as JSON field is deprecated")
                        data_formatted.append(
                            [
                                item["id"],
                                datetime.strptime(item["time"], "%Y-%m-%dT%H:%M:%S.%fZ"),
                                item["coordinate"][0],
                                item["coordinate"][1],
                                payload,
                            ]
                        )
                    df_single_entity = pd.DataFrame(
                        data_formatted, columns=["entity_id", "timestamp", "longitude", "latitude", "payload"]
                    )

                    if len(df_single_entity) > 0:
                        df_entities.append(df_single_entity)

            # Pandas is deprecating concatentation of empty dataframes, so we avoid adding empty
            # dataframes and check if the entire list of dataframes has data
            if len(df_entities) > 0:
                return pd.concat(df_entities).reset_index(drop=True)
            else:
                return pd.DataFrame(columns=["entity_id", "timestamp", "longitude", "latitude", "payload"])

        streamed_data = []

        if isinstance(polygon[0], (list, tuple)):
            # polygon passed by VERTICES - time and entity filters goes in the payload
            payload = {}
            if time_interval is not None:
                payload["startTime"], payload["endTime"] = unpack_and_convert_time_interval(time_interval)
            if entities is not None:
                payload["include"] = entities

            if filter_fields:
                payload["fields"] = query_doc.get_document()["filters"]["fields"]

            payload["vertices"] = polygon
            with self.conn.api_post("polygon/history", params=params, payload=payload) as response:
                streamed_data = self._receive("history", response)

        if isinstance(polygon[0], float):
            # polygon passed as a BOUNDING BOX - time and entity filters goes in the parameters
            if entities is not None:
                params["include"] = entities
            if time_interval is not None:
                params["startTime"], params["endTime"] = unpack_and_convert_time_interval(time_interval)

            min_lng, min_lat, max_lng, max_lat = polygon
            params.update(query_doc.filter_fields_url_encoding())
            with self.conn.api_get(
                f"bounding-box/{min_lng}/{min_lat}/{max_lng}/{max_lat}/history", params=params
            ) as response:
                streamed_data = self._receive("history", response)

        data_formatted = []
        for item in tqdm(streamed_data, disable=not self.conn.progress_bar, maxinterval=0.5, miniters=1):
            payload = {}
            if add_payload_as_json:
                if add_payload_as_json:
                    _logger.warning("Payload as JSON field is deprecated")
            data_formatted.append(
                [
                    item["id"],
                    datetime.strptime(item["time"], "%Y-%m-%dT%H:%M:%S.%fZ"),
                    item["coordinate"][0],
                    item["coordinate"][1],
                    payload,
                ]
            )
        return pd.DataFrame(data_formatted, columns=["entity_id", "timestamp", "longitude", "latitude", "payload"])

    def entities(
        self,
        dataset_id: str,
        polygon: Optional[models.Polygon] = None,
        time_interval: Optional[models.TimeInterval] = None,
        filter_fields: Optional[dict] = None,
    ) -> List[Union[str, int]]:
        """
        Get the list of entity ids within a space and optional time constraint.

        If time constraints are not passed, the whole dataset is returned.
        Start time and end time passed in the time_interval must be both valid datetime, with
        start_time < end_time or both None.

        :param dataset_id: the dataset to be queried.
        :param polygon: List of vertices [[lon1, lat1], [lon2, lat2], ...] or a list of four
            floats representing the bounding box extremes as [lon_min, lat_min, lon_max, lat_max].
            Non valid input will raise an error.
        :param time_interval: Tuple with the Lower bound and the upper bound time constraints.
        :param filter_fields: Optional[dict]: {field_name: {FilterOperation: value}}
        :returns: List of unique entities in time interval and polygon.
        :raises `DFIInputValueError`: If `time_interval` or `polygon` are ill-formed.
        :example:
        ```python
        from dfi import Client

        dfi = Client(token, url)

        dataset_id = "1234"
        start_time = datetime.strptime("2022-01-01T08:00:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ")
        end_time = datetime.strptime("2022-01-01T08:30:00.000Z", "%Y-%m-%dT%H:%M:%S.%fZ")
        pimlico_tube_station =  [
            [-0.13410870522207574, 51.48932327401289],
            [-0.1355577905032419, 51.48887697878598],
            [-0.1339996342869938, 51.487266398812636],
            [-0.13279985400049554, 51.487508959676234],
            [-0.13230124401115972, 51.4875768764868],
            [-0.1319428680814667, 51.487955268294115],
            [-0.1322389177630896, 51.4883918703234],
            [-0.1322389177630896, 51.48887697878598],
            [-0.13301799587057417, 51.48930386996324],
            [-0.13410870522207574, 51.48932327401289],
        ]

        dfi.get.entities(
            dataset_id,
            polygon = pimlico_tube_station,
            time_interval = (start_time, end_time),
        )

        ```
        """

        query_doc = _init_query_document(
            dataset_id=dataset_id, polygon=polygon, time_interval=time_interval, filter_fields=filter_fields
        )

        params = {"instance": query_doc.get_dataset_id()}
        payload = {}

        if polygon is None:
            if time_interval is not None:
                payload["startTime"], payload["endTime"] = (query_doc.get_min_time(), query_doc.get_max_time())
            params.update(query_doc.filter_fields_url_encoding())
            with self.conn.api_get("entities", params=params, stream=True) as response:
                return self._receive("entities", response)

        if isinstance(polygon[0], list) or isinstance(polygon[0], tuple):
            # polygon passed by VERTICES
            if time_interval is not None:
                payload["startTime"], payload["endTime"] = (query_doc.get_min_time(), query_doc.get_max_time())

            payload["vertices"] = query_doc.get_polygon()

            if filter_fields:
                payload["fields"] = query_doc.get_document()["filters"]["fields"]

            with self.conn.api_post("polygon/entities", params=params, payload=payload) as response:
                return self._receive("entities", response)

        if isinstance(polygon[0], float):
            # polygon passed as a BOUNDING BOX
            params = {"instance": dataset_id}
            if time_interval is not None:
                params["startTime"], params["endTime"] = unpack_and_convert_time_interval(time_interval)
            min_lng, min_lat, max_lng, max_lat = polygon
            params.update(query_doc.filter_fields_url_encoding())
            with self.conn.api_get(
                f"bounding-box/{min_lng}/{min_lat}/{max_lng}/{max_lat}/entities", params=params, stream=True
            ) as response:
                return self._receive("entities", response)

    def _receive(self, response_type: str, response: requests.models.Response) -> Union[List[Any], int]:
        """
        Helper function to parse clients events as entities, and optionally show the progress bar.
        Client events are submitted as sequence of events with a parity check embedded in the last one.

        1. the event list is empty → raise response error - empty list

        2. the event list contains no finish message → raise response error - no final event reached

        3. the event list does contain a finish message, but it gives a different number than then
        number of messages event in the event list → raise response error - events missing

        4. The event list does contain a finish message and tells a number of messages corresponding
        to the number of message received → return the combined messages content as result.
        If no messages, return the correct empty data structure, based on the query type.

        :param response_type: str in the list of available options. It will raise a KeyError if the one provided
        is not defined.
        :param response: requests.models.Response for the given query.
        """

        dict_response_type_init = {
            "entities": [],
            "history": [],
            "count": 0,
        }

        try:
            # different initialization according to what the user want to query.
            results = dict_response_type_init[response_type]
        except KeyError as err:
            raise KeyError(f"Allowed options for what to receive are {dict_response_type_init.keys()}") from err

        client = sseclient.SSEClient(response)

        events_list_is_empty = True
        events_list_has_no_finish_message = True
        num_messages = 0

        for event in (pbar := tqdm(client.events(), disable=not self.conn.progress_bar, maxinterval=0.5, miniters=1)):
            events_list_is_empty = False

            if event.event == "keepAlive":
                continue

            elif event.event == "message":
                num_messages += 1
                # different concatenation according to the response type.
                if response_type == "entities":
                    results += [json.loads(event.data)]
                elif response_type in ["history", "count"]:
                    results += json.loads(event.data)
                else:
                    raise KeyError()

                # if self.conn.progress_bar:
                pbar.set_description(f"Collecting {_len_of(results):,} records")
                continue

            elif event.event == "finish":
                events_list_has_no_finish_message = False
                dict_data = json.loads(event.data)
                if "messageCount" not in dict_data:
                    _logger.warning("Please make sure the DFI API is up to date and higher than 1.6.0")
                else:
                    num_events_message_declared = dict_data.get("messageCount")
                    if num_messages != num_events_message_declared:  # 3
                        _raise_events_list_discrepancy(num_messages, num_events_message_declared)
                break

            elif event.event == "queryError":
                _raise_query_error_event(event)

            else:
                _raise_unexpected_event_found(event)

        if events_list_is_empty:  # 1
            _raise_events_list_empty()

        if events_list_has_no_finish_message:  # 2
            _raise_finish_event_not_reached()

        return results  # 4


def _len_of(list_or_int: Union[List, int]) -> int:
    if isinstance(list_or_int, int):
        return list_or_int
    else:
        return len(list_or_int)


def unpack_and_convert_time_interval(time_interval: models.TimeInterval) -> Tuple[str, str]:
    """from time interval to start_time, end_time in isoformat."""
    # The user can give None to one of the two values here.
    # The time interval validation manages the available options.
    start_time, end_time = time_interval
    if start_time is not None:
        start_time = start_time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
    if end_time is not None:
        end_time = end_time.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
    return start_time, end_time


def _raise_query_error_event(event: sseclient.Event) -> None:
    """helper to raise a QueryError response error"""
    msg = f"event returned QueryError: {event.data}"
    _logger.error(msg)
    raise validate.DFIResponseError(msg)


def _raise_unexpected_event_found(event: sseclient.Event) -> None:
    """helper to raise a an unexpected event was found"""
    msg = f"Unexpected event in bagging area: {event}"
    _logger.error(msg)
    raise validate.DFIResponseError(msg)


def _raise_events_list_empty() -> None:
    """helper to raise an error if the event list was empty."""
    msg = "DFI provided an empty events list."
    _logger.error(msg)
    raise validate.DFIResponseError(msg)


def _raise_events_list_discrepancy(num_event_messages_recorded: int, num_event_messages_declared: int) -> None:
    """
    Helper to raise an error if the number of messages-events does not correspond to the
    number of messages recorded in the finishing message.
    """
    msg = f"""
    DFI provided an event list with {num_event_messages_recorded} events of type 'message',
    but finish message declared {num_event_messages_declared} events of type 'message' were passed.
    """
    _logger.error(msg)
    raise validate.DFIResponseError(msg)


def _raise_finish_event_not_reached() -> None:
    """helper to raise a warning if the "finish" event was not reached."""
    msg = "DFI provided no 'finish' event in the event list."
    _logger.error(msg)
    raise validate.DFIResponseError(msg)


def _init_query_document(
    dataset_id: str,
    polygon: Optional[models.Polygon] = None,
    time_interval: Optional[models.TimeInterval] = None,
    filter_fields: Optional[dict] = None,
) -> QueryDocument:
    """Shim for initalising a query document to pass into other calls"""

    query_doc = QueryDocument()
    query_doc.set_dataset_id(dataset_id)
    if time_interval:
        query_doc.set_time_range(time_interval=time_interval)
    if polygon:
        query_doc.set_polygon(polygon)
    if filter_fields:
        query_doc.add_filter_fields_from_dict(fields_dict=filter_fields)

    return query_doc
